// Allan Bowhill

#include "Disk.h"


Disk::Disk() : FO(NULL), entry(NULL)
	{
	}

// This gets called by session object first
// Here we initialize mechanism to build Categories, Comments and Links

void Disk::initialize(WebTree * wt)
	{
	WT = wt;
	Disk::readDisk();
	Disk::makeObjects();
	}

void Disk::readDisk()
	{
	// setup a list of traversable dirs to send to fts
	char * paths[2];			// declare parameter array
	const char * root = WT->getRootName();	// assign root path to const
	paths[0] = (char *)root;		// assign zero'th element as root
	paths[1] = '\0';			// assign first element as terminator
	char ** pathArray = paths;		// designate pointer for this list

	// open an fts(3) directory tree handle
	// arg 1: pointer to list of paths to parse
	// arg 2: use logical fs, ingore device files
	// arg 3: no sort order

	FO = fts_open(pathArray, FTS_LOGICAL | FTS_NOSTAT | FTS_XDEV, NULL);

	// If open failed, return error type with perror(3), then abort
	if (FO == NULL)
		{
		perror("Error in FTS open at EMP_ROOT");
		cout << ERROR_ABORT << endl;
		exit(1);
		}

	// call fts_read(3) on fts directory stream
	// exit with critical error if something is wrong with
	// entry or inode is missing from EMP root directory

	lastEntry = entry = fts_read(FO);

	if (entry == NULL && errno != 0 || entry->fts_ino == 0)
		{
		perror("Error in initial FTS read: No EMP_ROOT!");
		cout << ERROR_ABORT << endl;
		exit(1);
		}
	}

// Transfer Disk objects into Categories and Commentary objects via WebTree interface

void Disk::makeObjects()
	{
	Category * thisCategory = NULL;
	bool isRoot = true;
	status dirStatus;

	// walk entire directory heirarchy under root
	// only process pre-order directories

	while (entry != NULL)
		{

		// issue errors for entry
		testType(entry);

		// respond to directory entries
		if (entry->fts_info == FTS_D)
			{
			// obtain directory status
			dirStatus = getDirStatus(FO);

			// if there are categories or links in this dir, create Category object
			if (dirStatus.has_categories || dirStatus.has_links)
				{
				// The following three lines are absolutely required.
				// They create space on the heap for the string and null terminate it.

				int len = strlen(entry->fts_name);
				char * aDir = new (nothrow) char [len+1];
				assert(aDir);
				strlcpy(aDir, entry->fts_name, len+1);

				len = strlen(entry->fts_accpath);
				char * aPath = new (nothrow) char [len+1];
				assert(aPath);
				strlcpy(aPath, entry->fts_accpath, len+1);

				// add the category to the web tree
				thisCategory = WT->addCategory(aPath, aDir, entry->fts_level, dirStatus, NULL);

				// If this is first entry, it is the root category
				if (isRoot == true)
					{
					WT->addRoot(thisCategory);
					isRoot = false;
					}
				}
			// if emp root has files, but none are called "links", exit with an error
			else if (isRoot == true)
				{
				perror("No links file in emp root directory");
				exit(1);
				}
			}

		// process a commentary file
		else if (entry->fts_info == FTS_F)
			{
			// The following three lines are required to store 
			// the terminated string on the heap.

			int len = strlen(entry->fts_name);
			char * aFile = new (nothrow) char [len+1];
			assert(aFile);
			strlcpy(aFile, entry->fts_name, len+1);

			// add filename to webtree
			if (isCommentary(entry))
				if (dirStatus.has_comments && dirStatus.has_links)
					WT->addCommentary(thisCategory, aFile);
			}

		// process the next fts entry in tree
		entry = fts_read(FO);
		}

	// if errno != 0 we have an uncaught critical error in reading
	assert (errno == 0);

	// Close fts, handle errors

	if(fts_close(FO) != 0)
		{
		perror("fts_close error"); // looks up message and prints
		abort();
		}

	}


void Disk::transfer()
	{
	
	}


/*
Directory status is obtained through a call to fts_children(). The existence of
the following files are checked:

	1.) all directories
	2.) 'links'
	3.) *.commentary

A code is generated that fits an entry in the following table:

Directory Status Table                               Create object?

+------------+------------+------------------+ +----------------------------+
| categories | links file | commentary files | |Category   Link   Commentary|
+------------+------------+------------------+ +----------------------------+
|    yes     |   yes      |      yes         |     OK       OK       OK        Full category
|    yes     |   yes      |      no          |     OK       OK       NO        Cat & Links
|    yes     |   no       |      yes         |     OK       NO       NO        Just subcats
|    yes     |   no       |      no          |     OK       NO       NO        Just subcats
|    no      |   yes      |      yes         |     OK       OK       OK        Links & comment
|    no      |   yes      |      no          |     OK       OK       NO        Links only
|    no      |   no       |      yes         |     NO       NO       NO        ignore
|    no      |   no       |      no          |     NO       NO       NO        ignore
+------------+------------+------------------+

*/

status Disk::getDirStatus(FTS * FO)
	{
	status dirStatus;

	// reset flags
	dirStatus.has_links = dirStatus.has_categories = dirStatus.has_comments = 0;

	// init linked list to all files in dir, of which entry is first
	FTSENT * ent = fts_children(FO, NULL);

	// log and return if there were errors in fts_children

	if (ent == NULL && errno == 0)
		{
		}
	
	// Read all files and dirs in this directory
	// Set directory status bit flags to "on" for various conditions

	while (ent != NULL)
		{
		// if all flags are on, don't bother trying to throw any

		if (dirStatus.has_comments && dirStatus.has_links && dirStatus.has_categories)
			break;

		switch(ent->fts_info)
			{
			// set flag for categories

			case FTS_D:
				dirStatus.has_categories = 1;
				break;

			// set flag(s) for links and commentaries

			case FTS_F:
				if (strcmp(ent->fts_name, LINKS_FILE) == 0)
					dirStatus.has_links = 1;

				if (isCommentary(ent))
					dirStatus.has_comments = 1;
				break;
			}

		// make next link current to travel the linked list
		ent = ent->fts_link;
		}
	return dirStatus;
	}



/*
Process an entry from fts_children. 
Create an empty category object for each new directory. Assign name and a new parameter value.
Read links file, if one exists. Create new, populated links objects for each link.


*/
void Disk::testType(FTSENT * entry)
	{
	/*	
	// These are some diagnostic displays
	cout << endl;
	cout << "access path: " << entry->fts_accpath << endl;
	cout << "root path: " << entry->fts_path << endl;
	cout << "root path length: " << entry->fts_pathlen << endl;
	cout << "file name: " << entry->fts_name << endl;
	cout << "file name length: " << entry->fts_namelen << endl;
	cout << "directory level: " << entry->fts_level << endl;
	cout << "parent dir: " << entry->fts_parent << endl;
	cout << "stat  info: " << entry->fts_statp << endl;
	*/

	/* 
	Two of the following case statements (FTS_DC and FTS_F) catch file 
	and directory types that should be converted to categories, links, or 
	commentary objects. The rest are ignorable, non-fatal errors that are 
	logged as warnings.
	*/

	switch(entry->fts_info)
		{
		case FTS_D:
			// directory pre-order
			break;

		case FTS_DC:
			if (DEBUG > 1)
				cerr << "Error: "
				<< "A directory that causes a cycle in the tree" << endl
				<< "Circular directory: " << entry->fts_path << entry->fts_name << endl
				<< "Linked to: " << entry->fts_cycle->fts_path 
				<< entry->fts_cycle->fts_name << endl;
			break;

		case FTS_DEFAULT:
			if (DEBUG > 1)
			cerr << "Error: Not described" << endl;
			break;

		case FTS_DNR:
			if (DEBUG > 1)
			cerr << "Error: Directory cannot be read" << endl
			<< strerror(entry->fts_errno) << endl;
			break;

		case FTS_DOT:
			if (DEBUG > 1)
			cerr << "Error: Dot file" << endl;
			break;

		case FTS_DP:
			// cerr << "Directory post-order" << endl;
			break;

		case FTS_ERR:
			if (DEBUG > 1)
			cerr << "Error #: " << strerror(entry->fts_errno) << endl;
			break;

		case FTS_F:
			// Regular file
			break;

		case FTS_NS:
			if (DEBUG > 1)
			cerr << "Error: No stat info" << endl
			<< strerror(entry->fts_errno) << endl;
			break;

		case FTS_NSOK:
			if (DEBUG > 1)
			cerr << "Error: No stat info requested" << endl;
			break;

		case FTS_SL:
			if (DEBUG > 1)
			cerr << "Error: Symlink" << endl;
			break;

		case FTS_SLNONE:
			if (DEBUG > 1)
			cerr << "Error: Orphaned symlink" << endl;
			break;

		default:
			if (DEBUG > 1)
			cerr << "Error: Undefined FTS Error." << endl;
			
		}
	}

// search entry for Commentary suffix, and return true if found

bool Disk::isCommentary(FTSENT * ent)
	{
	char * tmp;

	if ((tmp = strstr(ent->fts_name, COMMENT_SUFFIX)) != NULL)
		if (strcmp(tmp, COMMENT_SUFFIX) == 0)
			return true;
	return false;
	}


