// Allan Bowhill

#include "Parameter.h"


Parameter::Parameter(): queryString(NULL), qlen(0)
	{
	// Buffer CGI query response string. 
	// Null-terminate it at 100 characters.

	queryString = new (nothrow) char [100];
	assert(queryString);

	char * q = getenv(QUERY_STRING);
	if(q != NULL)
		strlcpy(queryString, q, 100);

	// Get environment from webserver.

	httpUserAgent = getenv(HTTP_USER_AGENT);
	serverName = getenv(SERVER_NAME);
	serverAddr = getenv(SERVER_ADDR);
	remoteAddr = getenv(REMOTE_ADDR);
	scriptName = getenv(SCRIPT_NAME);

	// Check any CGI parameter recieved for 
	// validity. This is done by disassembling
	// the parameter field-by-field and checking
	// each field against an existing entity in
	// the WebTree. This way, we know the user
	// has requested a legitimate entity in 
	// the context of the request.
	//
	// If the fields pass, the parameter stack
	// is populated with the fields from the parameter
	// string recieved from the CGI.
	//
	// Reset paramter stack to NULL if CGI params
	// are invalid. This is equivalent to 
	// forcing the user request to the root
	// of the webtree. 

	}

// return CGI parameters to caller

int * Parameter::getParameterField()
	{
	int * anInt = (int *)Stack::peek();
	Stack::pop();
	return anInt;
	}


// returns: success (true) or failure (false)

bool Parameter::check()
	{
	// check the parameters for security
	if (Parameter::securityCheck() == false)
		return false;

	// then process the string and put items onto stack
	else if (Parameter::processString()== false)
		return false;

	else
		return true;
	}


// uses for input: queryString, queryPrefix
// output: queryString with queryPrefix removed
// returns: bool success (true) or failure (false)

bool Parameter::securityCheck()
	{
	// Get query prefix (default "q=") from header file and get its size.
	// We want to remove the prefix from the parameter string so there
	// are only numerals and parameter field seperators left.

	char queryPrefix [] = QUERY_PREFIX;
	int qClen = strlen(queryPrefix);

	// Test #1: Check for valid query prefix
	// Find first occurance of query prefix in parameter string.
	// If it is absent or not the first part of the query string, return false. 
	// the default query string preifix is "q=". Therefore a bad prefix would be:
	// QUERY_STRING="badprefix=1.0.1"
	// Otherwise, if the query string prefix is good, modify the query string 
	// to exclude the prefix.

	char * matchPtr = strstr(queryString, queryPrefix);
	if ((matchPtr == NULL) || ((matchPtr - queryString) != 0))
		return false;
	else
		queryString = matchPtr + qClen;

	// Test #2: Check for minimum request length
	// If the length of query string is less than three characters
	// (e.g: "1.0", "2.3", "4.5" are among the shortest), return false. 
	// Otherwise, continue checking the string in more detail for validity.

	qlen = strlen(queryString);
	if (qlen < 3)
		return false;


	// Test #3: Check ends of request string for context validity
	// If the first or last character are not digits, (e.g. - "a.10.0.1" or "8.9.") 
	// return false.

	if (!isdigit(queryString[qlen-1]) || !isdigit(queryString[0]))
		return false;

	// Test #4: Check to see if request is composed of valid characters
	// (e.g. - "." or numeric)
	// Test #5: Check minimum and maximum lengths of numeric fields
	// (e.g. ".." or fields over 3 digits long are rejects

	int digitCount = 0;
	int pcount = 0;

	for (int i = qlen-1; i > -1; --i)
		{
		if (isdigit(queryString[i]))
			{
			++digitCount;
			pcount = 0;
			}

		else if (queryString[i]== PARAM_FIELD_SEP)
			{
			++pcount;
			digitCount = 0;
			}
		else
			return false;

		if (pcount > 1 || digitCount > 3)
			return false;
		}

	return true;
	}


// input: queryString with queryPrefix removed
// output: integers onto stack representing values between the parameter field seperators 
// returns: bool success (true) or failure (false)

bool Parameter::processString(void)
	{
	// Process queryString in reverse order.
	// Convert numeric character data residing between the parameter field seperator dots
	// into single integers, and store them on the stack.
	// Although we may still have a valid but badly-formed parameter:
	// e.g. - "9.001.3", this function transforms it into: "9.1.3".

	// Initialize work variables.

	int dpos = 0;		// Digit position of number between seperators (1-3 digits.)
	int cumulative = 0;	// Integer translation of 3-digit numeric string.
	int ascdigit = 0;	// Ascii value of character.
	int powerofTen = 1;	// holds nth power of ten for conversions

	for (int i = qlen-1; i > -1; --i)
		{

		// if this character is not a parameter field seperator

		if (queryString[i] != PARAM_FIELD_SEP)
			{
			// place char in buf, and convert to integer
			char buf[2] = {'\0'};
			strlcpy(buf, &queryString[i], 2);
			ascdigit = atoi(buf);

			// if, for some reason, integer is not 0-9 return failure
			if (ascdigit < 0 || ascdigit > 9)
				return false;

			// Construct a single integer by accumulating value of each
			// converted char. We do this by working backwards through 
			// the group and adding increasing powers of 10 each time.

			else
				{
				cumulative += ascdigit * powerofTen;
				powerofTen *= 10;
				++dpos;
				}
			}
		else
			{
			// If we got here, we have either hit a parameter field seperator
			// or the end of the query string.


			// check cumulative value for a stand-alone zero
			// we should not have this, since EMP addresses all 
			// information groups > 0. This may change later, but
			// for now treat it as failure if this happens.

			if (cumulative == 0)
				return false;

			// Store integer on stack
			int * parm = new (nothrow) int[1];
			assert(parm);
			*parm = cumulative;
			Stack::push(parm);	

			// reset variables
			ascdigit = cumulative = dpos = 0;
			powerofTen  = 1;
			}

		}
	// Store last accumulated integer on stack
	int * parm = new (nothrow) int[1];
	assert(parm);
	*parm = cumulative;
	Stack::push(parm);	

	return true;
	}

