// Allan Bowhill

#include "Session.h"

// Constuctor
// 
// Initialize a new WebTree using EMP root as head node
//	Transfer category and link data from disk into webtree
// Initialize a new Parameter object to process CGI request 

Session::Session(DataSource * ds): aDataSource(ds)
	{
	aParameter = new (nothrow)Parameter;
	assert(aParameter);

	aConfig = new (nothrow)Config;
	assert(aConfig);

	// get the emp root dir
	empRoot = aConfig->getEmpRoot();

	if (empRoot == NULL)
		{
		perror("Missing EMP_ROOT");
		exit (1);
		}

	// initialize webtree object 
	WT = new (nothrow) WebTree(empRoot);
	assert(WT);

	aDataSource->initialize(WT);
	aDataSource->transfer();
	}

// travel the tree to see if an object exists for each field in 
// the parameter sent by the user

bool Session::processRequest()
	{
	int param;		// holds one field of CGI parameter
	bool match = false;	// if all fields of request are legit, set to true
	bool is_first = true;	// this flag allows us to skip first value on stack

	// fetch root category object
	Category * aCat = WT->getRoot();

	// check recieved parameters for legitimacy. If they are wrong in form,
	// delete them, and supply the root category as if it were the requested
	// item.

	if(aParameter->check() == false)
		{
		aParameter->clearStack();
		WT->setRequestedRoot(WT->getRoot());
		}

	// If the parameters appear legitimate, use them to determine if they
	// really point to an object in tree. If the object is exists, set the 
	// requested root pointer in WebTree to it.
	//
	// If no object is found, make the requested root the WebTree root.
	
	else while((param = *(aParameter->getParameterField())) != NULL)
		{
		// skip the first parameter, for it is always 1
		if (is_first == true)
			{
			is_first = false;
			continue;
			}

		if ((aCat = aCat->getSubCategory(param)) != NULL)
			{
			WT->setRequestedRoot(aCat);
			match = true;
			}

		else 
			{
			aParameter->clearStack();
			WT->setRequestedRoot(WT->getRoot());
			match = false;
			}
		}
	makePostObjects();
	return match;
	}

// traverse categories
void Session::traverse(Category * aCat)
	{
	Category * bCat = NULL;

	while((bCat = aCat->popSubCategory()) != NULL)
		traverse(bCat);
	}

// tell WebTree to create Links objects if file is present

void Session::makePostObjects()
	{
	Category * aCat = WT->getRequestedRoot();

	if (aCat->getStatus().has_links)
		makeLinks();

	if (aCat->getStatus().has_comments)
		readCommentaryFiles();
	}

// This makes links objects, and stores them in the links queue 
// in the Category object.

void Session::makeLinks()
	{
	// get the requested document root
	Category * aCat = WT->getRequestedRoot();

	// get the path for cwd
	char * path = aCat->getPath();

	// the links file is usually called 'links'
	char * file = LINKS_FILE;

	// zero out char attributes for Link object
	// Url: internet address
	// clickText: clickable text to go to that address
	// shortDescription: a description of the link
	char * Url = NULL;
	char * clickText = NULL;
	char * shortDescription = NULL;

	// character work buffer #1
	char buf[LINE_MAX] = {'\0'};

	chdir(path);

	// insure file stream is open, handle errors

	ifstream infile;
	infile.open(file);
	if (!infile) 
		{
		if (DEBUG > 0)
			cerr << "Error: file cannot be opened" << endl; 
		exit(1);
		}
	assert(infile);


	// read through leading crlf's, and get first line if it exists

	while(!infile.eof() && (strlen(buf) < 1))
		infile.getline(buf,LINE_MAX);

	// if we are left with a CRLF, links file is empty, so return

	if (strlen(buf) < 1)
		{
		if (DEBUG > 1)
			cerr << "links file is empty" << endl;
		return;
		}

	// We have a line, but don't know if it is the Category short
	// description or the first field of the first record. We determine
	// this by looking at the next line. 
	//
	// If the next line in the file is a CRLF, then the first line we 
	// encountered was a short description, and we have yet to encounter 
	// the first record.
	//
	// If the next line is not  a CRLF, then the first line is the
	// first field in the first record, and this line is the second field
	// in the first record.
	//
	// Get next line into 2nd buffer.

	char buf2[LINE_MAX] = {'\0'};
	infile.getline(buf2,LINE_MAX);

	
	// if it is EOF, store short description in object and return
	
	if (infile.eof())
		{
		char * membuf = new (nothrow) char [sizeof(buf)+1];
		assert(membuf);
		strlcpy(membuf, buf, sizeof(buf)+1);
		aCat->setShortDescription(membuf);
		return;
		}


	// if it is a CRLF, first line was category description
	// so assign it, then zip through any extra CRLFs until
	// we get to the first line of the first record or EOF
	
	else if (strlen(buf2) < 1)
		{
		char * membuf = new (nothrow) char [sizeof(buf)+1];
		assert(membuf);
		strlcpy(membuf, buf, sizeof(buf)+1);
		aCat->setShortDescription(membuf);

		// read through crlf's but put into buf, not buf2
		// we do this to make sure buf represents the 
		// first line of first record.

		// reset buf
		buf[0] = '\0';

		// get first field of first record
		while(!infile.eof() && (strlen(buf) < 1))
			infile.getline(buf,LINE_MAX);

		if (strlen(buf) < 1)
			return;

		// get second field of first record
	
		buf2[0] = '\0';
		infile.getline(buf2,LINE_MAX);
		assert((!infile.eof()) && (strlen(buf2) > 1));

		}
	// now we are loaded with the first line of the first record
	// assign this buf to URL string, and do the same to the next 
	// two lines of this first record

	// assign Url to heap
	Url = new (nothrow) char [sizeof(buf)+1];
	assert(Url);
	strlcpy(Url, buf, sizeof(buf)+1);

	// assign clickText
	clickText = new (nothrow) char [sizeof(buf2)+1];
	assert(clickText);
	strlcpy(clickText, buf2, sizeof(buf2)+1);
			
	// assign shortDescription
	buf2[0] = '\0';
	infile.getline(buf,LINE_MAX);
	assert((!infile.eof()) && (strlen(buf) > 1));


	// search buffer for an inline commentary tag, and recieve back
	// new buffer containing short description and embedded HTML clickable 
	// text, if commentary file exists.
	//
	// an inline commentary tag takes the form:
	//
	// [CLICKTEXT:filename.comment]
	// 
	// CLICKTEXT can be any clickable text, like MORE, EXTRA
	// filename.comment is the filename of the commentary file
	// which resides in the category's directory


	processShortDescription(buf);

	shortDescription = new (nothrow) char [sizeof(buf)+1];
	assert(shortDescription);
	strlcpy(shortDescription, buf, sizeof(buf)+1);

	// add first link object
	WT->addLink(aCat, Url, clickText, shortDescription);

	// Now, get remaining records in a somewhat normal fashion
	
	// If we are still not at EOF, then regard the
	// rest of the lines in the file as three-line
	// records, each trio (a record) seperated by one or 
	// more crlf's:
	//
	// line 1: URL of link
	// line 2: clickable text for Url
	// line 3: short description of link
	//
	// Before each record is read, discard crlf lines
	// that lead-up to first field in record.

	while(!infile.eof())
		{
		char buf[LINE_MAX] = {'\0'};

		// fields in a link record
		Url = NULL;
		clickText = NULL;
		shortDescription = NULL;

		// read through crlf's between records
		while(!infile.eof() && (strlen(buf) < 1))
			infile.getline(buf,LINE_MAX);

		if (strlen(buf) < 1)
			{
			if (DEBUG > 2)
				cerr << "no more links" << endl;
			return;
			}

			// now store the buf in heap for posterity
			char * membuf = new (nothrow) char [sizeof(buf)+1];
			assert(membuf);
			strlcpy(membuf, buf, sizeof(buf)+1);

				Url = membuf;

		// now we read a trio of fields in a record
		for (int i=0; i < 2; ++i)
			{
			// store a field in the buf
			buf[0] = '\0';
			infile.getline(buf,LINE_MAX);
			assert(strlen(buf) > 0);

			// now store the buf in heap for posterity
			char * membuf = new (nothrow) char [sizeof(buf)+1];
			assert(membuf);
			strlcpy(membuf, buf, sizeof(buf)+1);

			// Next, store membuf in approriate Link object 
			// attribute: URL, clickText or shortDescription
			
			if (i == 0)
				clickText = membuf;
			else
				shortDescription = membuf;

			}

		// Use WebTree interface to create a new Link
		if (Url != NULL && clickText != NULL && shortDescription != NULL)
			WT->addLink(aCat, Url, clickText, shortDescription);

		}

	assert(infile.eof());
	infile.close();
	}




// Process short description definition in third field from record
// look for text in the form:
//
// [CLICKTEXT:filename.comment]
//
// Scan for this text in the string. 
// Check Category's commentary queue for filename.comment
// If filename exists, replace this macro with HTML clickable text
// If it doesn't exist, just remove the macro
// return the modified buffer

void Session::processShortDescription(char * buf)
	{
	
	}

void Session::readCommentaryFiles()
	{
	// cout << "Making comment" << endl;
	}

void Session::display()
	{
	Display Html;
	Html.print(WT,aParameter);
	}
